module Database where

import Core

import Control.Exception            (IOException)
import Data.Aeson                   (ToJSON)
import Data.Maybe                   (listToMaybe)
import Data.Pool
import qualified Data.Text          as T
import Database.PostgreSQL.Simple
import Database.PostgreSQL.Simple.ToField (ToField)
import Effectful
import Effectful.Dispatch.Dynamic
import Effectful.Error.Static       (Error)
import Effectful.Reader.Static
import Effectful.State.Static.Local (State, get, put, evalState)
import Servant hiding               ((:>), throwError)

type instance DispatchOf Database = 'Dynamic

type DatabaseExeEffects es = ( Reader AppEnv :> es 
                             , Error ServerError :> es
                             , IOE :> es
                             )

databaseInit 
  :: (Database :> es, Error ServerError :> es) => Eff es ()
databaseInit = send DatabaseInit

databaseRead
  :: (ToField a, Show a, Database :> es, Error ServerError :> es)
  => (Query, a) -> Eff es [User]
databaseRead = send . DatabaseRead

databaseRead_
  :: (Database :> es, Error ServerError :> es) => Query -> Eff es [User]
databaseRead_ = send . DatabaseRead_

databaseWrite
  :: (ToRow a, Show a, Database :> es, Error ServerError :> es)
  => (Query, a) -> Eff es ()
databaseWrite = send . DatabaseWrite

runDatabaseDebug :: DatabaseExeEffects es => Eff (Database : es) a -> Eff es a
runDatabaseDebug = interpret $ \_ -> \case
  DatabaseInit -> do
    liftIOE $ putStrLn "Mocked setup of the database"
  DatabaseRead (statement, values) -> do
    liftIOE $ putStrLn
       $ "Mocked a READ database operation with statement:\n"
      ++ show statement
      ++ " and values:\n"
      ++ show values
    pure []
  DatabaseRead_ statement -> do
    liftIOE $ putStrLn "Mocked a READ database operation on all users"
    pure []
  DatabaseWrite (statement, values) -> do
    liftIOE $ putStrLn
      $ "Mocked a WRITE database operation with a user named " ++ show values

runDatabaseIO :: DatabaseExeEffects es => Eff (Database : es) a -> Eff es a
runDatabaseIO = interpret $ \_ -> \case
  DatabaseInit                      -> init
  DatabaseRead  (statement, values) -> read  statement values
  DatabaseRead_  statement          -> read_ statement
  DatabaseWrite (statement, values) -> write statement values
  where
    init :: DatabaseExeEffects es => Eff es ()
    init = do
      AppEnv { pool } <- ask
      liftWithPool pool $ \conn -> do
        execute_ conn createUsersTable
        execute_ conn createLogsTable
        pure ()

    read :: (ToField a, FromRow b, DatabaseExeEffects es)
         => Query -> a -> Eff es [b]
    read statement values = do
      AppEnv { pool } <- ask
      liftWithPool pool $ \conn ->
        query conn statement (Only values)

    read_ :: (FromRow b, DatabaseExeEffects es) => Query -> Eff es [b]
    read_ statement = do
      AppEnv { pool } <- ask
      liftWithPool pool $ \conn ->
        query_ conn statement

    write :: (DatabaseExeEffects es, ToRow a) => Query -> a -> Eff es ()
    write statement values = do
      AppEnv { pool } <- ask
      liftWithPool pool $ \conn -> do
        execute conn statement values
        pure ()

    liftWithPool p f = liftIOE $ withResource p f

createConnectionPool :: IO (Pool Connection)
createConnectionPool = newPool $ defaultPoolConfig
  (connectPostgreSQL "host=localhost dbname=demo")
  close
  60
  10

queryUser :: UserId -> (Query, UserId)
queryUser userId = ("SELECT id, name FROM users WHERE id = ?;", userId)

writeUser :: T.Text -> (Query, Only T.Text)
writeUser name = ("INSERT INTO users (name) VALUES (?);", Only name)

deleteUser :: UserId -> (Query, UserId)
deleteUser userId = ("DELETE FROM users WHERE id = ?;", userId)

createUsersTable :: Query
createUsersTable = "CREATE TABLE IF NOT EXISTS users ( \
                   \  id SERIAL PRIMARY KEY, \
                   \  name character varying(255) NOT NULL \
                   \);"

createLogsTable :: Query
createLogsTable = "CREATE TABLE IF NOT EXISTS logs ( \
                  \  id SERIAL PRIMARY KEY, \
                  \  level VARCHAR(10) NOT NULL, \
                  \  message TEXT NOT NULL, \
                  \  timestamp TIMESTAMPTZ NOT NULL DEFAULT NOW(), \
                  \  source VARCHAR(100), \
                  \  context JSONB \
                  \);"
